__all__ = ['GetDesignatedByRoleMethod',
           'RoleType'
           ]

from boa3.model.builtin.interop.role.getdesignatedbyrolemethod import GetDesignatedByRoleMethod
from boa3.model.builtin.interop.role.roletype import RoleType
