__all__ = ['IteratorMethod',
           'IteratorType']

from boa3.model.builtin.interop.iterator.iteratorinitmethod import IteratorMethod
from boa3.model.builtin.interop.iterator.iteratortype import IteratorType
